/*****************************************************************************
 * Filename:    cryptfs/cryptfs.c                                            *
 * Description: The main routine                                             *
 * Copyright:   2007 by Alexander Motzkau                                    *
 *****************************************************************************/
#include <getopt.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>

#include "actions.h"
#include "misc.h"
#include "xml.h"
#include "tty.h"

#define DEFAULT_XML "/etc/cryptfs.xml"

struct actionname
{
    char* name;
    struct actionname* next;
};

struct main_options
{
    char* xmlfile;
    struct actionname* actions;
    int ignore;
    enum on_mode mode;
    enum bool noenc;
};

void free_main_options(struct main_options* opts)
{
    struct actionname* ac;
    
    xfree(opts->xmlfile);

    ac = opts->actions;
    while(ac!=NULL)
    {
	struct actionname* temp;
	temp = ac;
	ac = ac->next;
	xfree(temp->name);
	xfree(temp);
    };
    
    xfree(opts);
};

struct main_options* parse_options(int argc, char* argv[])
{
    struct main_options* result;
    
    result = (struct main_options*)xmalloc(sizeof(struct main_options));
    result->xmlfile = NULL;
    result->actions = NULL;
    result->ignore = 0;
    result->mode = om_always;
    result->noenc = false;
    
    while(1)
    {
	int o;
	static struct option long_opts[] =
	    {
		{ "filename", 1, NULL, 'f'},
		{ "ignore", 0, NULL, 'i'},
		{ "only-map", 0, NULL, 'm'},
		{ "only-mount", 0, NULL, 'n'},
		{ "syslog", 0, NULL, 's'},
		{ "help", 0, NULL, 'h'},
		{ "no-enc", 0, NULL, 'e'},
		{ NULL, 0, NULL, 0 },
	    };
	    
	o = getopt_long(argc, argv, "f:ishmen", long_opts, NULL);
	if(o == -1)
	    break;
	
	switch(o)
	{
	    case '?':
		free_main_options(result);
		return NULL;
	    case 'f':
		if(result->xmlfile != NULL)
		{
		    errmsg("Filename given twice.\n");
		    free_main_options(result);
		    return NULL;
		}
		result->xmlfile = xmallocstr(optarg);
		break;
	    case 'i':
		result->ignore = 1;
		break;
	    case 'm':
		if(result->mode == on_mount)
		    result->mode = om_always;
		else
		    result->mode = on_map;
		break;
	    case 'n':
		if(result->mode == on_map)
		    result->mode = om_always;
		else
		    result->mode = on_mount;
		break;
	    case 'e':
		result->noenc = true;
		break;
	    case 's':
		use_syslog = 1;
		break;
	    case 'h':
		printf("Usage: cryptfs [options] [action ...]\n\n"
		       "Options:\n"
		       "  -f, --filename=XMLFILE  File containing the definitions.\n"
		       "                          Default: /etc/cryptfs.xml\n"
		       "  -i, --ignore            Ignore unknown actions.\n"
		       "  -m, --only-map          Don't fsck and mount.\n"
		       "  -n, --only-mount        Do everything after mapping.\n"
		       "  -e, --no-enc            Don't encrypt, just map the area.\n"
		       "  -s, --syslog            Error messages to syslog.\n"
		       "  -h, --help              Show options.\n");
		free_main_options(result);
		return NULL;
	}
    }
    
    if(result->xmlfile==NULL)
	result->xmlfile = xmallocstr(DEFAULT_XML);

    while(optind < argc)
    {
	struct actionname* action;
	action = (struct actionname*) xmalloc(sizeof(struct actionname));
	
	action->name = xmallocstr(argv[--argc]);
	action->next = result->actions;
	result->actions = action;
    }
    
    return result;
};

int main(int argc, char* argv[])
{
    struct configuration *config;
    struct main_options* opts;
    struct actionname* ac;
    struct actionname* pac;

    reserve_tty();
    setlocale(LC_CTYPE, "");
        
    opts = parse_options(argc, argv);
    if(opts == NULL)
	return 1;
        
    config = parse_file(opts->xmlfile);
    if(config == NULL)
	return 1;

    if(!complete_config(config))
	return 1;

    config->noenc = opts->noenc;
    if(config->noenc)
	config->mode = on_map;
    else
	config->mode = opts->mode;

    if(opts->actions == NULL) /* No action given. */
    {
	struct action *act;
	for(act = config->actions; act != NULL; act = act->next)
	    printf("%s\n", act->name);
	return 0;
    }

    for(pac = NULL, ac = opts->actions; ac != NULL; )
    {
	if(find_action(ac->name, config)==NULL)
	{
	    if(opts->ignore)
	    {
		if(pac == NULL)
		{
		    opts->actions = ac->next;
		    xfree(ac);
		    ac = opts->actions;
		}
		else
		{
		    pac->next = ac->next;
		    xfree(ac);
		    ac = pac->next;
		}

		continue;
	    }
	    else
	    {
		errmsg("Action '%s' undefined.\n", ac->name);
		return 1;
	    }
	}
	
	pac = ac;
	ac = ac->next;
    }
    
    for(ac = opts->actions; ac != NULL; ac = ac->next)
	if(!handle_action(ac->name, NULL, config))
	    errmsg("Action '%s' failed.\n", ac->name);

    release_tty();
        
    return 0;
}
